#include <Wire.h>
#include <Adafruit_GFX.h>
#include "Adafruit_LEDBackpack.h"
#include <driver/adc.h>

/*Time zone adjustment define the GPS module spits out time in UTC so
   we need to adjust it manually by a few hours. In my case it's -4 hours
   for EST */
#define TZAdjust -4

//want a 12 hour clock?
const boolean use12Hour = true;

//want to have the clock automatically adjust to daylight savings? 
const boolean doDaylightSavings = true;

#define LED1_BLUE 13
#define LED1_GREEN 12
#define LED1_RED 14

#define LED2_BLUE 27
#define LED2_GREEN 26
#define LED2_RED 25

Adafruit_7segment  matrix = Adafruit_7segment();

float rotationCounter = 0.1;
String inputData = ""; //stores a buffer of each line recieved by the GPS module, we only parse one of these
String timeString = ""; //stores the parsed time string for later processing
String dateString = ""; //stores the parsed date, this is used for adjusting daylights savings
byte ledMaxBrightness = 255;
int colorCounter = 0;
float ledFade = 1.0;
float dtFade = 0.08;
/* So basically i soldered the seven segment display in upside down because it was easier structurally
    as a result the bitmasks for numbers needed to be redefined since the adafruit library doesn't seem to
    support flipping the display, not a big deal but that's why this array exists. */
const uint16_t bitmasks[] = {
  0x003F, //0
  0x0030, //1
  0x005B, //2
  0x0079, //3
  0x0074, //4
  0x006D, //5
  0x006F, //6
  0x0038, //7
  0x007F, //8
  0x007C  //9
};

//updates the brightness with readings from the photoresistor, this just prevents the display from
//being super bright when the lights in the room are off. 
void updateBrightness() {
  //read the ADC attached between the photo resistor and 10k resistor 
  int read_raw;
  adc1_config_channel_atten((adc1_channel_t) 4, ADC_ATTEN_11db );
  read_raw = adc1_get_raw((adc1_channel_t) 4);

  //map the value to a brightness of 0 to 15 (0 is dim but not off) 
  matrix.setBrightness(constrain(map(read_raw, 700, 3800, 0, 15), 0, 15));

  //also edit the brightness of the RGB leds while we're at it
  ledMaxBrightness = constrain(map(read_raw, 700, 3800, 127, 255), 127, 255);
}

void setup() {
  Serial.begin(115200);
  Serial.println("GPS Clock Circuit Sculpture");

  // second serial channel, this one is used to read data from the GPS module
  Serial2.begin(9600, SERIAL_8N1, 16,  17);

  //configure PWM for the LEDs with 8-bit resolution
  for (int a = 0; a < 6; a++) {
    ledcSetup(a, 5000, 8);
  }

  //attach each of the pwm pins with their define
  ledcAttachPin(LED1_BLUE, 0);
  ledcAttachPin(LED1_GREEN, 1);
  ledcAttachPin(LED1_RED, 2);

  ledcAttachPin(LED2_BLUE, 3);
  ledcAttachPin(LED2_GREEN, 4);
  ledcAttachPin(LED2_RED, 5);

  //begin 7-seg display
  matrix.begin(0x70);
}

//Draws a number to the seven segment display (since the display is upside down i had to redefine the bitmasks instead of using the adafruit library)
void drawValue(uint16_t val) {
  matrix.writeDigitRaw(0, bitmasks[val % 10]);            //1's digit of the value
  matrix.writeDigitRaw(1, bitmasks[(val % 100) / 10]);    //10's digit of the value
  matrix.writeDigitRaw(2, 0x0003);                        //this one writes the colon in the middle
  matrix.writeDigitRaw(3, bitmasks[(val % 1000) / 100]);  //100's digit of the value
  matrix.writeDigitRaw(4, bitmasks[(val % 10000) / 1000]);//1000's digit of the value

  //optional debug print to spam the terminal with
  //  Serial.println("Value: " + String(val) + " Bitmasks: " + bitmasks[(val % 10000) / 1000] + " " + String(bitmasks[(val % 1000) / 100], HEX) + " " + String(bitmasks[(val % 100) / 10], HEX) + " " + String(bitmasks[val % 10], HEX));
}

//splits data by seperator and allows us to parse data out of ordered strings.
//this function came from stack overflow post: https://stackoverflow.com/questions/29671455/how-to-split-a-string-using-a-specific-delimiter-in-arduino
//and it just works fantastic so i keep reusing it
String getValue(String data, char separator, int index)
{
  int found = 0;
  int strIndex[] = {0, -1};
  int maxIndex = data.length() - 1;

  for (int i = 0; i <= maxIndex && found <= index; i++)
  {
    if (data.charAt(i) == separator || i == maxIndex)
    {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i + 1 : i;
    }
  }

  return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}

//handles the fading animation of the LEDs
void updateLEDs() {
  for (int a = 0; a < 3; a++) {
    //color counter just selects which color is currently on.
    //this is the only place the value is used. it is updated below
    //here the ESP32 only lights 1 color at a time 
    byte brightness = (colorCounter % 3 == a) ? ledMaxBrightness * ledFade : 0;

    //write values to the LED, the channel declarations are set up so we can light up 
    //both LEDS the same color with these two statements
    ledcWrite(a, brightness);
    ledcWrite(a + 3, brightness);

    //terminal spam
    Serial.println("Channels " + String(a) + " " + String(a + 3) + " Set to " + brightness);
  }

  //we use the fade variable to control the current brightness of the LEDs, it ranges from 0 to 1.
  ledFade += dtFade;

  //hit 1 start counting down
  if (ledFade > 1) {
    dtFade = - dtFade;
    ledFade = 1;
  }

  //hit 0 start counting up and switch to the next color
  if (ledFade < 0) {
    dtFade = - dtFade;
    ledFade = dtFade;
    colorCounter++;
  }

  //terminal spam
  Serial.println("Fade Set to " + String(ledFade, 3));
}

void loop() {

  //update brightness and LEDs
  updateBrightness() ;
  updateLEDs();

  // write the changes we just made to the display
  matrix.writeDisplay();
  delay(75);

  //read data from the GPS module 
  while (Serial2.available()) {
    char currentChar = Serial2.read();
    Serial.print(currentChar);

    //if the character is a new line we analyze the data coming in to see
    //whether it contains our time information
    if (currentChar == '\n') {

      //Looking for the GPRMC NMEA sentence, technically we could use GPGGA also
      //but updating the time every second is more than
      //good enough for a clock that doesnt actually have any second indicator.
      if (getValue(inputData, ',', 0).equals("$GPRMC")) {
        Serial.println("Parsing Time Data....");
        timeString = getValue(inputData, ',', 1);

        //terminal spam
        Serial.println("Parsed Time Data: " + timeString);

        //now we have the time data as a string, the format is HHMMSS so it's just a matter of parsing that out
        int hour = timeString.substring(0, 2).toInt();
        int minute = timeString.substring(2, 4).toInt();

        //adjust to timezone
        hour += TZAdjust;

        //daylights savings code
        //here we parse the date out of the GPS module datastream and determine whether we're 
        //in the -1 hour time of the year
        if (doDaylightSavings) {

          //parse out the raw date string
          dateString = getValue(inputData, ',', 9);
          Serial.print("Parsed Date Information: " + dateString);

          //extract meaningful values
          int day, month;
          day = dateString.substring(0, 2).toInt();
          month = dateString.substring(2, 4).toInt();

          Serial.print("    Current Day: " + String(day) + " Month: " + String(month));

          //-1 hour between november first and march 8th
          if ((month > 10 || month < 3) || (month == 3 && day < 8)) {
            hour -= 1;
            Serial.println("  Daylight savings enabled -1 Hour");
          }

        }

        //hours can't really be negative here, if we got negative numbers its
        //because of timezone adjust and daylights savings so we are actually just
        //on the previous day
        if (hour < 0) {
          hour = 24 + hour;
        }

        //converting time to 12 hour clock
        if (use12Hour && hour > 12) {
          hour -= 12;

          //12 hour clocks don't really use 0 in the hour place, so if we're using that clock format
          //then 0 hours is actually 12
          if (hour == 0) {
            hour = 12;
          }
        }

        Serial.println("Current Hour: " + String(hour) + " Current Minute: " + String(minute));

        //write new time value to the display
        drawValue(hour * 100 + minute);
      }

      //get rid of the input data we parsed from to make room for the next one
      inputData = "";
    } else {

      //not a new line character that was received so just add it to input data
      //string so we can analyze the entire sentence later
      inputData += currentChar;
    }
  }
}
